function pfold = computePfoldWrapper(transitionProbsFilename, ...
				     foldedIndicesFilename, ...
				     unfoldedIndicesFilename,...
				     startIndex, endIndex)

% pfold computation for a protein..
% returns empty if (there are > 1 connected components(cc)) and
% either 
% :) folded or unfolded indices fall into the smaller cc (not
% to the biggest cc). 
% or
% :) nodes startIndex..endIndex(both 1-based) don't fall into the biggest CC.
% returns the pfold for nodes startIndex..endIndex. 
% These indices refer to the original
% indices, and are properly adjusted in case of > 1 cc.

if (startIndex == 0)
  fprintf(1, 'error! startIndex should be 1-based\n');
  exit;
end

if ((startIndex > endIndex) | (startIndex < 0))
  fprintf(1, 'error! check indices into computePfoldWrapper\n');
  exit;
end

transitionProbs    = load (transitionProbsFilename);


tmp1               = max(transitionProbs(:,1)) + 1;
tmp2               = max(transitionProbs(:,2)) + 1;

numNodes           = max(tmp1, tmp2);

nodesInBiggestCC   = findBiggestCC(transitionProbs);
foldedIndices      = load (foldedIndicesFilename);
unfoldedIndices    = load (unfoldedIndicesFilename);
Tsparse            = prepareTsparse(transitionProbs);
returnedIndices    = startIndex:endIndex;

if (length(nodesInBiggestCC) ~= numNodes)
  
  %now updating folded and unfolded macrostate indices
  
  origLengthFolded    = length(foldedIndices);
  origLengthUnfolded  = length(unfoldedIndices);
  
  foldedIndices       = findNewIndices(nodesInBiggestCC, foldedIndices);
  unfoldedIndices     = findNewIndices(nodesInBiggestCC, unfoldedIndices);
  
  newLengthFolded     = length(foldedIndices);
  newLengthUnfolded   = length(unfoldedIndices);
  
  if ((newLengthFolded ~=  origLengthFolded) | ...
      (newLengthUnfolded ~= origLengthUnfolded))
    fprintf(['discarding roadmap, since some folded or unfolded nodes' ...
	     ' are not in biggest CC\n']);

    pfold = [];
%    printPfold(pfold);
    return;
  end
  
  returnedIndices    = findNewIndices(nodesInBiggestCC, startIndex:endIndex);
  if (length(returnedIndices) ~= endIndex - startIndex + 1)
    fprintf(['discarding roadmap, since some of the test pfold nodes' ...
	     ' are not in biggest CC\n']);
    pfold = [];
    return;
  end
  
  %updating Tsparse
  Tsparse            = Tsparse(nodesInBiggestCC, nodesInBiggestCC);
end

%[numNodes, ignore] = size(Tsparse);
%keyboard;
allpfold           = computePfold(Tsparse, foldedIndices, unfoldedIndices, ...
				  1);
pfold              = full(allpfold(returnedIndices));

%printPfold(pfold);
  

%function printPfold(pfold)

%fid   = fopen('pfold.txt','w');
%fprintf(fid, '%roadmap of %d nodes but considered the biggest CC  of %d nodes\n', numNodes, length(nodesInBiggestCC));
%fprintf(fid, '%f\n', pfold');
%fclose (fid);

