//#include "indices.H"
#include <stdlib.h>
#include "protein.H"
#ifndef _DEBUG_H
  #include "debug.H"
#endif  
//#include "simplex.H"
#include <time.h>
#include "tree.H"
#include <vector>
#include <fstream>
using std::ofstream;
#include "MCSimulation.H"
#include <math.h>
#include <iostream>
using std::cout;
using std::endl;

extern const float  INV_BOLTZMAN_K  = 1.677481737;
//extern const float  INV_BOLTZMAN_K  = 0.1677481737;
bool                PRINT           = false;

/* 
   used exclusively by roulette. 
   Example : 
   input    array : 1 2 3 4.5        sum =10.5
   step 1   array : 9.5 8.5 7.5 6 (= {10.5 -1, 10.5 - 2, 10.5 - 3, 10.5 -4 })
   returned Array : 9.5 18 25.5 31.5 

*/
void printPath(const vector<int>& path, Protein ** allNodes, int runIndex);

/*
  prints non-loop centroid coords to the file MCResult_CentroidCoords{runIndex}.txt.
  The file has the following format:
  # nodes on the path
  # centroid coords to write
  -----
  energy
  centroid # 1 coords
  centroid # 2 coords
  centroid # 3 coords
  ...
  centroid # N coords
  -----
  energy
  centroid # 1 coords
  ...
  eof
*/
void printPath(const vector<int>& path, Protein ** allNodes, int runIndex)
{
  unsigned int i;
  char atomCoordsFilename[100];

  sprintf(atomCoordsFilename, "MCResult_CentroidCoords%d.txt", runIndex);

  /*
    ofstream fout ("MCResult_energy.txt");
    ofstream fout2("MCResult_rmsd.txt");
  */

  // just touching the output file
  ofstream fout3(atomCoordsFilename);
  fout3 << path.size() << endl;
  int numAtomsToWrite =  Protein::lastAAIndexOfLastNonLoopVector_plusOne- 
    Protein::firstAAIndexOfFirstNonLoopVector;
  for (int j = 0; j < numAtomsToWrite; j++)
    {
      if (Protein::loop[j + Protein::firstAAIndexOfFirstNonLoopVector])
	numAtomsToWrite --;
    }
  
  fout3 << numAtomsToWrite << endl;
  fout3.close();
  /*
    cout << "check MCResult_energy.txt " << endl;
    cout << "check MCResult_rmsd.txt " << endl;
  */
  cout  << "check " << atomCoordsFilename << endl;
  
  for (i = 0; i < path.size(); i++)
    {
      //      fout  << allNodes[path[i]]->getEnergy() << endl;
      //      fout2 << rmsd(allNodes[path[i]], allNodes[0]) << endl;
      allNodes[path[i]]->printEnergyAndCentroidCoords2File(atomCoordsFilename);
    }
  
  //  fout.close();
  //  fout2.close();
}

/* 
   selects the next node based onprobabilities. Returns the index of the
   selected next node. Returns # neighbors, if stays at same node.

*/

   
int roulette_VP (const vector<float>&  probabilities) 
{
  vector<float> cumProbs; // cumulative normalized probabilities

  double sigmaProb = 0.0, // sum of probabilities, to normalize.
    rN, stayProb;
  const    int UNINITIALIZED = -1;
  int          retval        = UNINITIALIZED;
  unsigned int i;

  // compute stay prob.
  
  for (i = 0; i < probabilities.size(); i++)
    {
      sigmaProb += probabilities[i];

      if (i != 0)
	cumProbs.push_back(probabilities[i] + cumProbs[i-1]);
      else
	cumProbs.push_back(probabilities[i]);
    }
  
  stayProb       = probabilities.size() - sigmaProb;
  cumProbs.push_back(probabilities.size());
  
  assert (fabs(cumProbs[cumProbs.size() - 1] - 
	       cumProbs[cumProbs.size() - 2] - 
	       stayProb) < 1E-3);

  // select a random number between 0 and n.
  rN       = drand48() * probabilities.size();
  
  // depending on where it falls, select the next node index.
  for (i = 0; i < cumProbs.size(); i++)
    {     
      if (rN <= cumProbs[i])
	{
	  retval = i;
	  break;
	}
    }

  assert(retval !=  UNINITIALIZED);
  
  return retval;
}

/* returns true according to Metropolis criterion */
bool metropolisAccept (float startEnergy, float endEnergy)
{
  if (endEnergy < startEnergy)
    return true;
  else
    {
      float rN = random() / MY_RAND_MAX;
      if (rN < exp (- (endEnergy - startEnergy) * INV_BOLTZMAN_K))
	{
	  return true;
	}
    }
  return false;
}

/*
  does 1 step of MC simulation. continues trying to find a config.
  until one is accepted. startEnergy is the energy corresponding to the
  startconfig, which is stored in p.

  returns newParams and energy, and creates the new config in p.
  also updates numIter.
*/
void MC_OneStep (const vector<float> & startParams, /*epsilon, cyclic*/
	 	 vector<float> & newParams, 
		 Protein *p, 
		 float startEnergy,
		 float * energy, 
		 unsigned int * numIter, 
		 const Protein * nativeP)
{
  const float   epsilon = 0.05;
  float         exclusionEnergy;



  assert       (fabs(p->getEnergy() - startEnergy) < 1E-1);
  
  //  p may not have params set, according to the new params.
  do 
    {
      do 
	{
	  newParams       = sampleAround         (startParams, epsilon,
						  Protein::cyclic);
	  /*
	    float dist = 0.0;
	    for (unsigned int i = 0; i < params.size(); i++)
	    {
	    dist += (params[i] - newParams[i]) * 
	    (params[i] - newParams[i]);
	    cout << "distance between params is : " << sqrt(dist) << endl;
	  */
	  p->createGivenConfig                   (newParams, nativeP);
	  /*
	    cout << "rmsd between startP and this protein is : ";
	    cout << rmsd(p, startP) << endl;
	  */
	  p->setEnergySet(false);
	  (*energy)         = p->computeEnergy();
	  exclusionEnergy   = p->getExclusionEnergy();
	  //cout << "energy of new config = " << (*energy) << endl;
	  //cout << "exclusion energy = " << exclusionEnergy << endl;
	  //	  numTrials ++; 
	}
      while (exclusionEnergy > Protein::maxExEnergy);
      //      numAccepted ++;
      (*numIter)++; // config does not self collide, increasing numIterations.
    }
  while (!metropolisAccept(startEnergy, (*energy)));
}

/*
  does numIter iterations, with an interval of printInterval.
  each generated conformation, even if it is not accepted by
  the Metropolis criterion, is counted in the MC simulation.
  starts from startP, and cyclic contains info. on whether the 
  parameters are cyclic.

  also ends when the config. has an rmsd of < 3A to the 
  folded state.
*/
void MCSimulate (Protein * startP, double numIter, double printInterval, 
		 const Protein * nativeP, const vector<bool> & cyclic,
		 Protein ** allNodes)
{
  cout << "fix MCSimulate." << endl;
  exit(0);
#if 0
  vector<float> newParams, params, fullparams;
  float         energy1, energy2, f_rmsd; 
  bool          foundFoldedState = false;
  unsigned int  i = 0;
  Protein *     p;
  
  cout << "inv. boltzmann const = " << INV_BOLTZMAN_K << endl;

  assert(startP->IsEnergySet());
  assert(startP->IsParamsSet());
  
  energy1         = startP->getEnergy();
  fullparams      = startP->getParams();
  
  //  cout << "startP has " << fullparams.size() << " params." << endl;
  params.resize(0);

  //   remove twist angle from params;

  //   0 1 2 3 4 5  6  7  8  9  10  11  12  13  14  15  16 17 18 19 20
  //   t         t                  t                      t
  //   twistangle is at 0, 5, 11, 17, 17 + 6, ... i.e. 5 + 6n.
  
  
  for (unsigned int j = 0; j < fullparams.size(); j++)
    {
      /*
	if ( (j == 0) || (j == 5) || ( (j > 5) && ((j - 5) %  6 == 0) ) )
	{
	// do nothing
	}
	else
	{
      */
      //	  cout << "pushing in " << j << " th parameter." << endl;
      cout << j << " th parameter is " << fullparams[j] << endl;
      params.push_back(fullparams[j]);
      //}
    }
 
  p = new Protein(-1);
  
  p->createGivenConfig(params, nativeP);
  p->computeEnergy    ();
  p->findParams       ();

  cout << "startP is : "      << (*startP) << endl;
  cout << "p->getEnergy() = " << p->getEnergy() << " energy1 = " << energy1 << endl;
  
  assert (fabs(p->getEnergy()      - energy1) < 1E-1);
  assert (rmsd(p, startP)              < 1E-1); 
  
  vector<float> debugParams = p->getParams();
  assert (debugParams.size() == fullparams.size());
  
  //cout << "in MCSimulate" << endl;
  
  for (unsigned int i = 0; i < params.size(); i++)
    {
      //cout << params[i] << " " <<  debugParams[i] << endl;
      assert (fabs(params[i] - debugParams[i]) < 1E-2);
    }
  
  //  int numAccepted = 0, numTrials = 0;
  
  while ((i < numIter) && (!foundFoldedState))
    {
      MC_OneStep(params, /*epsilon, cyclic*/newParams, p, energy1, &energy2, 
		 &i, allNodes[0]);
      
      f_rmsd     = rmsd(nativeP, p);
      
      if (f_rmsd < 3)
	{
	  foundFoldedState = true;
	}
      
      if ((i % (int)printInterval == 0) || (foundFoldedState))
	{
	  cout << "step #   " << i << endl;
	  cout << "energy = " << energy2 << " rmsd = ";
	  cout << f_rmsd  << endl;
	}
      
      energy1 = energy2;
      params  = newParams;
      //cout << "acceptance ratio = " << numAccepted * 1.0 / numTrials << endl;
      //cout << "energy = " << energy1 << endl;
    }

  delete p;
#endif
}

/*
  sets rmsd2Folded, rmsd2Unfolded; foundFoldedState and foundUnfoldedState.
  first call should have  minRMSDFolded, minRMSDUnfolded, foldedProtein and
  unfoldedProtein set to desired values and firstCall set to true. 
  Later calls should not set the last 3 parameters.
*/

void checkLocation(Protein * currentProtein, float *rmsd2Folded, 
		   float *rmsd2Unfolded, bool *foundFoldedState, 
		   bool * foundUnfoldedState,
		   bool firstCall = false,
		   float minRMSDFolded=0.0, float minRMSDUnfolded=0.0,
		   Protein * foldedProtein= NULL, 
		   Protein * unfoldedProtein= NULL)
{
  static float f_minRMSDFolded, f_minRMSDUnfolded;
  static Protein * p_foldedP, * p_unfoldedP;
  if (firstCall)
    {
      f_minRMSDFolded   = minRMSDFolded;
      f_minRMSDUnfolded = minRMSDUnfolded;
      p_foldedP         = foldedProtein;
      p_unfoldedP       = unfoldedProtein;
    }
  
  *rmsd2Folded   = p_foldedP  ->lsqfit_RMSD(currentProtein);
  *rmsd2Unfolded = p_unfoldedP->lsqfit_RMSD(currentProtein);

  if (*rmsd2Folded < f_minRMSDFolded)
    {
      *foundFoldedState = true;
    }
  
  if (*rmsd2Unfolded < f_minRMSDUnfolded)
    {
      *foundUnfoldedState = true;
    }
}

/*
  returns 0  if reached close to folded state.
          1                     unfolded state.
	 -1  if neither.
*/	  
int MCSimulate (int startNodeIndex, Protein ** allNodes, 
		const unsigned int maxIter,
	        int  foldedStateIndex, int unfoldedStateIndex, 
		float minRMSDFolded, float minRMSDUnfolded)
{
  cout << "fix MCSimulate." << endl;
  exit(0);
  return 0;
#if 0
  Protein *     p;
  vector<float> params, newParams;
  float         energy1, energy2, rmsd2Folded, rmsd2Unfolded;
  unsigned int  i = 0;
  bool          foundFoldedState, foundUnfoldedState;
  int           retval;
  
  assert   (allNodes[startNodeIndex]->IsEnergySet());
  assert   (allNodes[startNodeIndex]->IsParamsSet());

  energy1 = allNodes[startNodeIndex]->getEnergy();
  params  = allNodes[startNodeIndex]->getParams();
  
  p       = new Protein(-1);
  
  p->createGivenConfig(params, allNodes[0]);
  p->computeEnergy    ();
  p->findParams       ();

  assert (fabs(p->getEnergy()      - energy1)      < 1E-1);
  assert (rmsd(p, allNodes[startNodeIndex]) < 1E-1); 
  
  foundFoldedState = foundUnfoldedState = false;

  checkLocation (p, &rmsd2Folded, &rmsd2Unfolded, &foundFoldedState, 
		 &foundUnfoldedState, true, minRMSDFolded, minRMSDUnfolded,
		 allNodes[foldedStateIndex], allNodes[unfoldedStateIndex]);

  while ((i < maxIter) && (!foundFoldedState) && (!foundUnfoldedState))
    {
      MC_OneStep(params, newParams, p, energy1, &energy2, 
		 &i, allNodes[0]);
      
      checkLocation (p, &rmsd2Folded, &rmsd2Unfolded, 
		     &foundFoldedState, &foundUnfoldedState);
      
      if (i % 10000 < 20)
	{
	  cout << "iteration # " << i << endl;
	  cout << "rmsd2Folded = " << rmsd2Folded << " rmsd2Unfolded = ";
	  cout << rmsd2Unfolded << endl;
	}
      
      energy1 = energy2;
      params  = newParams;
    }

  delete p;

  // finding return value...
  
  if ((foundFoldedState) && (foundUnfoldedState))
    {
      cout << "starting from node # " << startNodeIndex << ", reached ";
      cout << "both folded and unfolded state. "        << endl;
      cout << "both folded and unfolded state. the distances are :";
      cout << "dist 2 folded state   = " << rmsd2Folded << endl;
      cout << "dist 2 unfolded state = " << rmsd2Unfolded << endl;
      retval = -1;
    }
  else if (foundUnfoldedState)
    {
      retval = 1;
      cout << "unfolded." << endl;
    }
  else if (foundFoldedState)
    {
      retval = 0;
      cout << "folded." << endl;
    }
  else
    {
      cout << "neither folded nor unfolded." << endl;
      retval   = -1;
    }

  cout << "did " << i << " iterations." << endl;
  return retval;
  #endif
}


/*
  does at most maxIter iterations in each simulation, and 
  maxNumMCSimulations simulations.
*/
void pFoldComputation (int foldedStateIndex, int unfoldedStateIndex, 
		       int numNodes, int startIndex,   
		       float minRMSDFolded,
		       float minRMSDUnfolded,
		       Protein ** allNodes)
  
{
  const unsigned int       maxIter          = (int) 1E6;
  int             folded, unfolded, simulationResult, numMCSimulations;
  const int       maxNumMCSimulations = 300; // max per node.
  char            filename[100];
  
  sprintf(filename, "pfold%d.txt", startIndex);
  cout << "check " << filename << endl;

  ofstream fout(filename);
  fout << "%pfold computation, " << minRMSDFolded << " and " << minRMSDUnfolded << " radii, " << maxNumMCSimulations << " sim./node, starting from node # " << startIndex << endl;
  fout << "%max " << maxIter << " iterations / node." << endl;
  fout << "%format: nodeIndex #folded #unfolded #numMCsimulations" << endl;

  cout << "starting from node # " << startIndex << endl;
  folded = unfolded = numMCSimulations = 0;
  
  while ( numMCSimulations < maxNumMCSimulations)
    {
      simulationResult = MCSimulate(startIndex, allNodes, maxIter, foldedStateIndex,
				    unfoldedStateIndex, minRMSDFolded, 
				    minRMSDUnfolded);
      if (simulationResult == 0)
	{
	  cout << "folded" << endl;
	  folded ++;
	}
      else if (simulationResult == 1)
	{
	  cout << "unfolded" << endl;
	  unfolded ++;
	}
      else
	{
	  cout << " not folded nor unfolded " << endl;
	}
      numMCSimulations++;
      
      //	  if ((finalNode == 0) || (finalNode == 1))
      //{
      //if ((folded + unfolded) % 100 == 0)
      if (numMCSimulations % 100 == 0)
	{
	  fout << startIndex << " " << folded << " " << unfolded << " ";
	  fout << numMCSimulations << endl;
	}
      //}
      cout << "node # " << startIndex << " " <<  (folded * 1.0/ numMCSimulations) << endl;
      //      fout << i << " " <<  (folded / (folded + unfolded + 0.0)) << endl;
    }

  fout.close();

}
