#ifndef _ANALYSIS_H
#define _ANALYSIS_H

#include "tree.H"
#include "protein.H"
//#include "pairOfVector.H"
#include <string>
#include <vector>
#include "matrix.H"

class Path;
class Protein;
struct neighbourStruct;
typedef struct neighbourStruct neighbourType;
/* 
   for the histogram, what is the percentage of elements passing close by of 
   the nodes of the best path.
   newPath is the pathIndex'th path that is stored. 
*/
class visitsClass {
public:

  visitsClass();
  int index;
  int numVisited;
  static int numNonZeroElements;
};

void countIndicesInNewPathClose2RootIndices (Path * newPath, 
					     Path * root, 
					     int pathIndex, int K,
					     Protein ** allNodes
					     /*float threshold*/);


/* i/p params : writes the file to ./protein_name/energyVsRMSD.txt.*/
void analyzeDistvsEnergy(char * protName, 
			 int numberOfNodes, Protein ** allNodes);

void analyzeRMSDvsEnergy(char * protName, 
			 int numberOfNodes, Protein ** allNodes);

void findNearOptimalPaths(int endNodeIndex, int numberOfNodes, 
			  Protein ** allNodes, int numPathsToReport);

bool findNearOptimalPathsPreliminaryCheckOK(int startSearchNodeIndex, 
					    int endSearchNodeIndex,
					    Protein ** allNodes);
void interactivePrintNodesInfo(Path *path,Protein ** allNodes);

/* interactively prints info about protein 
   whose indices are entered by the user. */

void printInteractiveNodeInfo(Protein ** allNodes);

void printMostVisitedNodeInfo(float *A,  Path *root, int numIndices, Protein ** allNodes);
void findBinIndicesOfProtein(int protIndex, int *a, int *b, Protein ** allNodes);
bool checkRange(float f, float lowerBound, float upperBound);
/* prints most visited elements, at most numNodes2Print of them. */
void printMostVisitedNodes(visitsClass * visits, int numberOfNodes, 
			   int numNodes2Print, Protein ** allNodes,
			   char * outFilename);

/* first sorts the array. Then, gathers close proteins together. Starts from the biggest cluster, and adds other samples close to it. But does not change the original cluster in this process. (Maybe it should ? How can I change the cluster?) */

void aggregateVisitedArray(visitsClass * visits, int numberOfNodes, 
			   Protein ** allNodes);



int inverse_sort_compare(const void *vfirst, const void *vsecond);

void initializeVisitsArray(visitsClass ** visits, int numberOfNodes);

void analyzeMinEnergyNodeConnections(int nodeIndex, Protein ** allNodes);
int  findMinEnergyNode  (Protein ** allNodes, int numberOfNodes); 

/*
  returns indices of the K nodes of energy < avg. energy/ 5.0.
  the # nodes returned can be found from indices.size().
  also reports min, max and avg energy of proteins.
*/

void findMinEnergyNodes (vector<int> & indices, int K, 
			 Protein ** allNodes, const int numNodes);

void printNodeEnergies (Protein ** allNodes, int numberOfNodes);

/*
  indices contain low energy node indices.
  For each of the nodes in indices, finds the avg.  path weight in and out, 
  and prints to the file.
*/
   
void analyzeAvgPathWeights (Protein ** allNodes, int numNodes, 
			    const vector<int> & indices,
			    char * outFilename);


/*
  for each seed, prints out the energy, distance from the goal,
  avg. path weight of all the paths entering it, and leaving it.
*/
   
void analyzeAvgPathWeights (const vector<neighbourType> & sampledAroundList,
			    Protein ** allNodes, int numRandomNodes,
			    int numNodes, 
			    const char * outFilename);

/* 
   finds the avg. incoming and outgoing path weights for
   node # index.
   returns them in aWI and aWO.
*/

void findAvgWeights (float * aWI, float * aWO, int index,  
		     int numRandomNodes,
		     int numNodes, Protein ** allNodes);

/* 
   finds and returns at max. K configs, which are at least R distant 
   (RMSD distance)
   from each other, and which have the lowest energy.
*/

void findMinEnergyClusters    (vector<int> & indices, unsigned int K, 
			       Protein ** allNodes,  int numNodes,
			       float R, int goalNodeIndex);
bool connected (int i, int j, Protein ** allNodes, int numNodes);
unsigned int findNumNeighbors (int i, int numNodes, Protein ** allNodes);

void obtainEdgeStats(int numNodes, Protein ** allNodes);
vector<int> computeCloseNodes(int centerNodeIndex, Protein ** allNodes, 
			      int numNodes,
			      float macrostateRMSDRadius);
void computeAndPrintPfold (vector<Real> & pfolds, int numPfoldsRequested, int maxNumNeighbors, 
			   int runIndex,
			   float foldedMacrostateRMSDRadius, 
			   float unfoldedMacrostateRMSDRadius, 
			   int numberOfNodes,
			   Protein**  allNodes,
			   int foldedStateIndex, int unfoldedStateIndex);

vector<int> computeCloseNodes(int centerNodeIndex, Protein ** allNodes, 
			      int numNodes,
			      float macrostateRMSDRadius);

void        findNodes2Expand (vector<int> & centerNodeIndices, int numNewNodes  , 
			      const vector<Real> & pfolds , Protein ** allNodes);
#endif

