#include "protein.H"
#include <math.h>
#include "debug.H"
#include <vector>
#include <iostream>
using std::cout;
using std::endl;

/*
  functions not normally used, but sometimes to be replaced
  with original functions for debugging purposes (e.g., instead of expensive
  energy computation, putting a simple computation)
  are kept here.
*/

// only checks pairwise node distances.
// also prints the max. distance encountered so far.
void checkProteinDistances (Protein ** allNodes, int numberOfNodes) {

  const float epsilon = 1E-3;
  float /*temp1, */temp2, temp3;
  int i;
  static float maxDist = 0.0;

  for (i = 0; i < numberOfNodes; i++) {
    //    temp1 = allNodes[i]->getRMSDFromNative();
    temp2 = allNodes[0]->computeDistance(allNodes[i]);
    if (temp2 > maxDist)
      {
	maxDist = temp2;
	//cout << "maxDist = " << maxDist << endl;
      }
    temp3 = allNodes[i]->computeDistance(allNodes[0]);

    /*
    if (fabs(temp1 - temp2) > epsilon) {
      cout << "temp1 = " << temp1 << " temp2 = " << temp2 << endl;
      cout << "i = " << i << endl;
      assert(0);
    }
    */
    if (fabs(temp2 - temp3) > epsilon) {
      cout << "temp3 = " << temp3 << " temp2 = " << temp2 << endl;
      cout << "i = " << i << endl;
      assert(0);
    }
    /*
    if (fabs(temp1 - temp3) > epsilon) {
      cout << "temp1 = " << temp1 << " temp3 = " << temp3 << endl;
      cout << "i = " << i << endl;
      assert(0);
    }
    */
  }
}
/* 
   This is a temporary debugging file aimed at checking whether every node in allNodes array has its index parameter properly set.
*/
void checkProteinIndices(Protein ** allNodes, int numberOfNodes) {

  int i;
  
  for (i = 0;  i < numberOfNodes; i++) {
    assert(allNodes[i] ->getIndex() == i);
  }
}


void checkProteinEnergies      (Protein ** allNodes, int numberOfNodes)
{
  int i;
  
  for (i = 0; i < numberOfNodes; i++) 
    {
      if (allNodes[i]->getExclusionEnergy() > Protein::maxExEnergy)
	{
	  cout << "i = " << i << "exclusion Energy = " << allNodes[i]->getExclusionEnergy() << " total energy = " << allNodes[i]->getEnergy() << endl;
	  assert(allNodes[i]->getExclusionEnergy() <= Protein::maxExEnergy);
	}
    }
}
void checkProteinParams(Protein ** allNodes, int numberOfNodes)
{
  unsigned int proteinNumDof = Protein::NUMDOF;
  for (int i = 0; i < numberOfNodes; i++)
    {
      assert(allNodes[i]->getParams().size() == proteinNumDof);
    }
}

void checkProteins      (Protein ** allNodes, int numberOfNodes)
{
  checkProteinIndices(allNodes, numberOfNodes);
  checkProteinDistances  (allNodes, numberOfNodes);
  checkProteinEnergies(allNodes, numberOfNodes);
  checkProteinParams(allNodes, numberOfNodes);
}

/* prints the tree name for each node. */
void checkTrees (Protein ** allNodes, int numNodes)
{
  for (int i = 0; i < numNodes; i++)
    {
      cout << "node # " << i << " tree name : " ;
      cout << allNodes[i]->myTree->myTreeName << endl;
    }
}
