#include "myVector.H"
#include "vec_utility.H"
#include <stdlib.h>
#include "protein.H"
#include <iostream>
using std::cerr;
using std::endl;
#include <string.h>
#include <map>
#include "sse.H"
#include "tree.H"
#include <time.h>
#include "tree.H"
#include "analysis.H"
#include "debug.H"
#include <fstream>
#include <sys/times.h>
#include "MCSimulation.H"
#include "ioRoutines.H"
#include <sys/stat.h> // for mkdir,access
#include <sys/types.h>//for getpid
#include "roadmapConstruction.H"
#include "matrix.H"
#include "unistd.h" // for getpid


/* extern variables */
extern Protein*      allNodes[];
extern int           i_MAXNODES;
/** 
    <b>vectorbased</b><br>
    <em>created on \today</em>
*/
/**@name predict.C */
//@{

/* global variables */
bool writeTransitionProbs = false,
  createNewRoadmap = false;

/// main function.
int main(int argc, char ** argv) 
{
  char relDir[100], protein_name[100], roadmapFilename[100];

  int           numberOfNodes = 0, 
    numRandomNodes,     extendedNodeIndex, 
    startIndex, //  the node u start MC simulation from for pfold computation.
    runIndex, // the run index of the MC simulation.
    maxNumNeighbors,
    nodeReadStartIndex = 2, nodeReadEndIndex = 80,
    numPfoldsRequested = 80,
    numNodesAroundEachCenter = 0;

  char readSomeNodesFromThisFile[] = "./tests/1ROP/myRoadmap10002.txt";
  
  vector<Real> pfolds;
  
  const int 
    foldedStateIndex = 0, unfoldedStateIndex = 1;

  //  vector<bool>           cyclic;
  unsigned int seed         =    time(NULL) * getpid();
  float foldedMacrostateRMSDRadius, unfoldedMacrostateRMSDRadius;
  
  cerr << "(using " << ((sizeof(Real)==sizeof(float))? "float": "double") << ")" << endl;

  cout << "seed = is random and equal to " << seed << endl;
  srandom             (seed);
  
  handleInput         (argc, argv, &numRandomNodes, relDir, protein_name,
		       roadmapFilename, &startIndex, &runIndex,
		       &createNewRoadmap, 
		       &foldedMacrostateRMSDRadius,
		       &unfoldedMacrostateRMSDRadius);
  //  assert (numRandomNodes == 100);
  cout << "folded macrostate radius   = " << foldedMacrostateRMSDRadius << endl;
  cout << "unfolded macrostate radius = " << unfoldedMacrostateRMSDRadius << endl;
  
  printParameters     (numRandomNodes, protein_name, 
		       createNewRoadmap);
  
  initNativeStructure (&numberOfNodes, relDir, argv[1], 
		       foldedStateIndex);
  
  //cout << "native state initialized. IT IS :" << endl;
  assert  (allNodes[0] != NULL);
  //cout << *allNodes[0] << endl;
  
  initExtendedStructure (&numberOfNodes, allNodes[0]);
  extendedNodeIndex   = numberOfNodes - 1;
  assert                (extendedNodeIndex  == 1);
      
  readSomeNodesFromFile  (readSomeNodesFromThisFile, allNodes,  
			  &numberOfNodes, nodeReadStartIndex, 
			  nodeReadEndIndex);
  createNodes        (&numberOfNodes,  numberOfNodes, numRandomNodes);
  numPfoldsRequested       = numberOfNodes;
  numNodesAroundEachCenter = Protein::NUMDOF;
  checkProteins      (allNodes, numberOfNodes);
  linkNodes          (numberOfNodes, runIndex, Protein::NUMDOF);  
  
  maxNumNeighbors    = (int) computeTransitionProbabilities(allNodes, numberOfNodes);
  

  computeAndPrintPfold ( pfolds, numPfoldsRequested,  maxNumNeighbors, 
			 runIndex,
			 foldedMacrostateRMSDRadius, 
			 unfoldedMacrostateRMSDRadius, 
			 numberOfNodes,
			 allNodes,
			 foldedStateIndex,  unfoldedStateIndex);

 
  return 0;
}

//@}
/*@Include: rmsd_nearestNeighbors.C roadmapConstruction.C ioRoutines.C */
/*@Include: analysis.C matrix.C tree.C protein.C */
