/* 
   to find nearest neighbors, using rmsd with rigid transformations.
*/


#include "protein.H"
#include <vector>
#include "sse.H"
#include <math.h>
#include <utility>
using std::pair;
#include <iostream>
using std::cout;
using std::endl;


vector<int>       nn_idx;			// near neighbor indices
vector<float>	  dists;			// near neighbor distances
pair<float,int>* distArray = NULL;

/**@name rmsd_nearestNeighbors.C */
//@{

///when relinking, creates a larger distArray.
void reinitializekNN(int numNodes)
{
  assert (nn_idx.size() == dists.size());
  assert (nn_idx.size() != 0);

  assert (distArray != NULL);
  delete[] distArray;
  distArray = new pair<float,int>[numNodes-1];
}

///initialize data structures to do the brute-force search of coefficient*k nearest neighbors.
void initializekNN(int numNodes, int k, int coefficient)
{
  cout << "Each node will be tried to be connected to " <<  k << " nodes , out of " <<  coefficient*k << " nodes." << endl;
 
  nn_idx.resize(coefficient*k);
  dists.resize (coefficient*k);
  distArray = new pair<float,int>[numNodes-1];
}

///for sorting pair of float,int's.
int floatCompare (const void* pv1, const void* pv2){
  pair<float,int> p1,p2;
  p1 = * ( (pair<float,int>*) pv1);
  p2 = * ( (pair<float,int>*) pv2);
  float f = p1.first - p2.first;
  if ( fabs(f) < 1e-7 )
    return 0;
  if (f < 0)
    return -1;
  else
    return 1;
}


/** 
   finds and returns k nearest nodes of a given node.
   does optimal Xformation before computing distance(rmsd).
   put the result in nn_idx, and dists arrays.
   These arrays should have been initialized already.
*/
void findNearestNeighbors(int nodeIndex, int numNodes, Protein * allNodes[], 
			  unsigned int k) 
{

  assert (distArray != NULL);

  

  for (int i=0,j=0;i<numNodes;i++,j++)
    {
      if (i!= nodeIndex)
	{
	  j = (i<nodeIndex) ? i : i-1 ; //we are skipping over i, so adjust index of distArray...

	  distArray[j].first  = allNodes[nodeIndex]->lsqfit_RMSD(allNodes[i]);
	  distArray[j].second = i;
	}
    }

  
  qsort(distArray, numNodes-1, sizeof(pair<float,int>), floatCompare);


  assert (nn_idx.size() == k);
  assert (dists.size()  == k);

  for (unsigned int i=0;i<k;i++)
    {
      nn_idx[i] = distArray[i].second;
      dists[i]  = distArray[i].first;

    }

}

//@}
