#include "sse.H"
#include <iostream>
using std::cout;
using std::endl;
#include "vec_utility.H"
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <vector>

extern Protein myProtein;

/* 
   this function does not compute bond angle and dihedral angles, so for
   native structure computebonddihedralangles, and for nonnative structures,
   setvector should be called to compute them.
*/

Sse::Sse (const sseType itype,const Point ibase, 
	  const Point iend,  int inumAAs) 
{

  float length;

  Point diff;

  type = itype;
  vec.setBase(ibase);
  vec.setEnd (iend);

  VecSub(diff,iend,ibase);
  length = VecLength(diff);
  
  
  vec.setLength(length);

  vectorAngle = dihedralAngle = 0.0;
  
#if 0
  if (length != 0.0) 
    {
      theta = acos (diff[2]/length);
      if ((diff[1] == 0.0) && (diff[0] == 0.0)) 
	{
	  phi = 0;
	}
      else 
	{
	  phi   = atan2(diff[1],diff[0]);
	}
      //    cout << "theta = " << theta << " phi = " << phi << " v = (" << diff[0] << "," << diff[1]<< "," << diff[2] << ")" << endl; 
    }
  else 
    {
      //cout << "Length == 0.0. Cannot compute theta and phi.." << endl;
    }
#endif
  numAAs     = inumAAs;
  twistAngle = 0.0;
}

/* 
   used by computeVector to find a coordinate of endPoint exclusively. 
*/

float findEndPointCoord (float iLength, float endPtCoord, float basePtCoord) 
{
  const float minusEpsilon = -0.005;

  float tempFloat   = iLength * iLength - 
    (endPtCoord - basePtCoord) * (endPtCoord - basePtCoord);

  if (tempFloat < minusEpsilon) {
    cout << "can't take the sqrt of a negative number." << endl;
    cout << "iLength = " << iLength << " endP[2] = " << endPtCoord << 
      "basePtCoord = " << basePtCoord << " total : " << 
      iLength * iLength - (endPtCoord - basePtCoord) * 
      (endPtCoord - basePtCoord) << endl;
    exit(0);
  }
  else if (tempFloat < 0) {
    return 0.0;
  }
  else return (sqrt(tempFloat));
  return 0.0;
}

void Sse::computeVector(const Point& ibase, float ilength, float vectorAngle,
			float dihedralAngle, Sse * prevSse, 
			Sse * prevprevSse)
{
  
  Point endPoint;

  if (prevSse == NULL) 
    {
      assert(prevprevSse == NULL);
      endPoint[0] = endPoint[1] = 0.0; endPoint[2] = ilength;
      this->vectorAngle =  this->dihedralAngle = 0.0;
    }
  else 
    {
      if (prevprevSse == NULL) 
	{
	  endPoint[0] = 0.0; 
	  endPoint[2] = ibase[2] - ilength * cos(vectorAngle);
	  endPoint[1] = findEndPointCoord (ilength, endPoint[2], ibase[2]);
	  this->vectorAngle        = vectorAngle; 
	  this->dihedralAngle      = 0.0;
	}
      else 
	{
	  obtainFourthPoint (prevprevSse->getBasePoint(), 
			     prevprevSse->getEndPoint(), 
			     prevSse->getEndPoint(), 
			     endPoint, dihedralAngle,    
			     vectorAngle, ilength);
	  
	  this->vectorAngle        = vectorAngle;    
	  this->dihedralAngle      = dihedralAngle;
	}
    }
  
  vec.setBase  (ibase);
  vec.setEnd   (endPoint);
  vec.setLength(ilength);
}

ostream& operator<< (ostream & output, Sse * s) 
{
  
  switch(s->type) 
    {
    case ALPHA : 
      {
	output << "structure: ALPHA" << endl;
	break;
      }
    case BETA : 
      {
	output << "structure: BETA" << endl;
	break;
      }
    case LOOP : 
      {
	output << "structure: LOOP" << endl;
	break;
      }
    default: 
      {
	output << "What is the type??" << endl;
      }
    }//switch
  
  output << "Num AAs        = " << s->numAAs << endl;
  output << s->vec;
  //  output << "theta = " << s->theta << " phi = " << s->phi << endl;
  output << "dihedral Angle = " << s->dihedralAngle << " vectorAngle = " << s->vectorAngle;
  output << "twistAngle     = " << s->twistAngle << endl;
  return output;
}

void Sse::setVector2(const Point base, const Point end) {
  
  Point diff;
  float length;

  vec.setBase(base);
  vec.setEnd(end);

  VecSub(diff, end, base);
  length = VecLength(diff);

  vec.setLength(length);
 
}

sseType Sse::getType() const {
  return type;
}

const Point& Sse::getBasePoint() const {
  return vec.getBase();
}

const Point& Sse::getEndPoint() const {
  return vec.getEnd();
}
#if 0
float Sse::getPhi() const {
  return phi;
}

float Sse::getTheta() const {
  return theta;
}
#endif

float Sse::getDihedralAngle() const {
  return dihedralAngle;
}

float Sse::getVectorAngle () const {
  return vectorAngle;
}

float Sse::getLength() const {
  return vec.getLength();
}
#if 0
void Sse:: computeEndPoint(const Point& base, float theta, float phi, float length, Point endPoint) {
  
  assert(length > 0);
  endPoint[0] = base[0] + length * sin(theta) * cos(phi);
  endPoint[1] = base[1] + length * sin(theta) * sin(phi);
  endPoint[2] = base[2] + length * cos(theta);
}
#endif

void Sse::incrementNumAAs( int increment) {

  numAAs += increment;
}

int Sse::getNumAAs() const {
  return numAAs;
}

void Sse::setTwistAngle(float f) {
  twistAngle = f;
}

float  Sse::getTwistAngle() const {
  return twistAngle;
}

void Sse::setVectorAngle(float b) {
  vectorAngle = b;
}

void Sse::setDihedralAngle(float d) {
  dihedralAngle = d;
}

/* 
   calls the other computeVector, and sets up the vector # sseIndex of protein.
*/
void Sse::computeVector(int sseIndex, float vectorAng, float dihedralAng,
			float twistAng, 
			float length, Protein * fatherProt) 
{
  const Point originPoint = {0.0, 0.0, 0.0};
  Point base;
  Sse * prevprevSse;
  Sse * prevSse;

  assert ((sseIndex >= 0) && (sseIndex < fatherProt->getNumSseElements()));

  if (sseIndex == 0) {
    prevSse = prevprevSse = NULL;
    VecCopy(base, originPoint);
  }
  else {
    VecCopy(base, fatherProt->getSseElement(sseIndex-1)->getEndPoint());
    prevSse     = fatherProt->getSseElement(sseIndex-1);

    if (sseIndex == 1) {
      prevprevSse = NULL;
    }
    else {
      prevprevSse = fatherProt->getSseElement(sseIndex - 2);
    }
  }

  computeVector (base, length, vectorAng, 
		 dihedralAng, prevSse, prevprevSse);
  setTwistAngle(twistAng);


}

/* 
   this function does not compute bond angle and dihedral angles, so
   for native structure computebonddihedralangles, and for nonnative 
   structures, setvector should be called to compute them. 

   ibase and iend have 3 floats in them.
*/

Sse::Sse (const sseType itype, const vector<float> & ibase, 
	  const vector<float> & iend,  int inumAAs) 
{

  float length;
  Point p1, p2;

  assert (ibase.size() == 3);
  assert (iend.size() == 3);

  p1[0] = ibase[0];  p1[1] = ibase[1];  p1[2] = ibase[2];
  p2[0] = iend[0];   p2[1] = iend[1];   p2[2] = iend[2];
  
  Point diff;

  type = itype;
  vec.setBase(p1);
  vec.setEnd (p2);
  
  VecSub(diff,p2,p1);
  length = VecLength(diff);
  
  vec.setLength(length);

  vectorAngle = dihedralAngle = 0.0;
  
  numAAs     = inumAAs;
  twistAngle = 0.0;
}
