#include "tree.H"
#include "analysis.H"
#include <math.h>
#include <vector>
#include <sys/times.h>
#include <time.h>
#include <iostream>
using std::cout;
using std::endl;

extern void initNodeColor(const int numNodes,  int thecolor, 
			  Protein ** allNodes) ;
extern void initQue(queType ** myQue, float t, nodeClass *fromN, nodeClass *toN);
extern void runBFS(queType * myQue, nodeClass *targetnode);
extern void printIntArray(int * A, int numElements);
extern float rmsd(const  Protein * p1, const Protein * p2);


/**@name tree.C */
//@{

bool noCommonNodeBetween(Path * p1, Path * p2);

int Path::startIndex = -1, Path::endIndex = -1, Path::numElementsInTree = 0;
const int Path::uninitializedDeviatedEdgeIndexValue = -2;

treeClass::treeClass() {

  root     = NULL;
  nextTree = NULL;
  prevTree = NULL;
  numNodesInTree = 0;

}

void treeClass::printTree() 
{
  static int totalNumNodes = 0;
  static int printed = 0;

  cout << "Tree name: " << myTreeName << " Tree Size: " << numNodesInTree << endl;
  //  cout << "indices in the tree : " << endl;
  //  printNodeIndices ();
  totalNumNodes += numNodesInTree;
  
  if (nextTree != NULL) 
    {
      nextTree->printTree();
    }
  
  if (!printed)
    {
      cout << "total num nodes = " << totalNumNodes << endl;
      printed = 1;
    }
}

/**
   moves the nodes in the (this) connected component 
   to replacementTree connected component. updates 
   properly globalFirstTree, globalLastTree, 
   and replacementTree.
   
   replacementTree can be NULL, in which case 
   the nodes myTree pointer is set to 
   NULL, and the other parameters are also taken ca
 */
void treeClass::removeTree (int * numTrees, treeClass* replacementTree,
			    treeClass ** globalFirstTree, 
			    treeClass ** globalLastTree, 
			    Protein   ** allNodes, 
			    const int    numNodes)
{
  if (replacementTree != NULL)
    replacementTree->numNodesInTree += numNodesInTree;

  initNodeColor( numNodes,0 , allNodes);
  
  root->replaceTreePtr(replacementTree, NULL);
  
  if (prevTree != NULL) 
    {
      prevTree->nextTree = nextTree;
    }
  else 
    {
      (*globalFirstTree) = nextTree;
    }

  if (nextTree != NULL) 
    {
      nextTree->prevTree = prevTree;
    }
  else 
    {
      (*globalLastTree) = prevTree;
    }
  
  (*numTrees)--;
  
}

//holds links (edge) information. 
ptrList::ptrList() {

  nextPtr    = NULL;
  node       = NULL;
  pathWeight = 0.0;
  edgeTransitionProbability = 0.0;
}

nodeClass::nodeClass() {

  firstLink = NULL; 
  lastLink  = NULL; 
  myConfig  = NULL; 
  pathPtr   = NULL;
  nodeColor = (char)(0); 

}

///removes all edges going out of this node, resets its edge parameters accordingly.
void nodeClass::removeAllLinks()
{
  ptrList * tempEdge, * edge = firstLink;
  
  while (edge != NULL)
    {
      tempEdge = edge;
      edge     = edge->nextPtr;
      delete tempEdge;
    }
  firstLink = lastLink = NULL;
  pathPtr   = NULL;  // this may not be necessary.
  nodeColor = (char)(0); 
  myConfig->setNumLinks(0);
}

///adds link from (this) to node "thelink" with weight "wt" to the end of the links from current node.
void nodeClass::addNodeLink(nodeClass *thelink, float wt) 
{

  ptrList *tempPtrList;

  tempPtrList             = new ptrList;

  tempPtrList->node       = thelink;
  tempPtrList->pathWeight = wt;

  if (lastLink == NULL) 
    {
      lastLink  = tempPtrList;
      firstLink = tempPtrList;
    }
  else 
    {
      lastLink->nextPtr = tempPtrList;
      lastLink          = tempPtrList;
    }
}

/** 
    changes the connected component link of all the nodes in the same connected component,
    to point to the connected component pointed by replacement tree.
*/
void nodeClass::replaceTreePtr (treeClass *replacementTree, 
				nodeClass *fromNode) 
{
  ptrList *linkPtr;

  myConfig->myTree = replacementTree;
  nodeColor        = (char)(1);
  linkPtr          = firstLink;
  
  while (linkPtr  != NULL) 
    {
      if ((linkPtr->node != fromNode) && 
	  (linkPtr->node->nodeColor == (char)(0)))
	{
	  linkPtr->node->replaceTreePtr(replacementTree, this);
	}
      linkPtr = linkPtr->nextPtr;
    }
}

void addToQue(queType * myQue, float t, nodeClass *fromN, nodeClass *toN) {

  queType *tempQuePtr, *prevQuePtr, *newQuePtr;
  int done;
 
  tempQuePtr = myQue;
  prevQuePtr = NULL;

  done = 0;
  while ((tempQuePtr != NULL) && (done == 0)) {
    if (tempQuePtr->timeStamp <= t) {
      prevQuePtr = tempQuePtr;
      tempQuePtr = tempQuePtr->nextQEntry;
    }
    else {
      done = 1;
    }
  }

  newQuePtr = new queType;

  newQuePtr->timeStamp = t;
  newQuePtr->fromNode  = fromN;
  newQuePtr->toNode    = toN;

  newQuePtr->nextQEntry = tempQuePtr;
  prevQuePtr->nextQEntry = newQuePtr;
}


void nodeClass::setPath(queType * myQue,nodeClass *prevNode, 
			float cTime) 
{

  ptrList *linkPtr;

  pathPtr = prevNode;
  nodeTime = cTime;
  
  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node->myConfig->getIndex() == 0) {
    }
    if (linkPtr->node->nodeColor == (char)(0)) {
      addToQue(myQue, cTime + linkPtr->pathWeight, this, linkPtr->node);
      if (linkPtr->node->myConfig->getIndex() == 0) {
      }
    }
    linkPtr = linkPtr->nextPtr;
  }

}

void nodeClass::setPath_reverse(queType * myQue, nodeClass *prevNode, 
				float cTime) 
{

  ptrList *linkPtr;

  pathPtr = prevNode;
  nodeTime = cTime;
  

  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node->nodeColor == (char)(0)) {     
      addToQue(myQue, cTime + (linkPtr->node->reverseWeight(this)), this, 
	       linkPtr->node);
    }
    linkPtr = linkPtr->nextPtr;
  }
}
 
float nodeClass::reverseWeight(nodeClass *prevNode) {

  ptrList *linkPtr;
  float retval = -1.0;

  linkPtr = firstLink;

  while (linkPtr != NULL) {
    if (linkPtr->node == prevNode) {
      retval = linkPtr->pathWeight;
      break;
    }
    linkPtr = linkPtr->nextPtr;
  }

  if (retval < 0) {
    assert(retval == -1.0);
    cout << "path not found " << endl;
    retval = 0.1;
  }

  return retval;  
}


void deleteQueue(queType * myQue) {
 
  queType * tempQuePtr, *prevQuePtr;

  tempQuePtr = myQue;

  while (tempQuePtr != NULL) {
    prevQuePtr = tempQuePtr;
    tempQuePtr = tempQuePtr -> nextQEntry;;
    delete prevQuePtr;
  }
}

/* 
   array[0..numElements - 1] should contain the elements to be reversed, again in array.
 */
void reverseArray(int * array, int numElements) {

  int i, temp;
  
  
  for (i = 0;  i < (numElements / 2); i++) {
    temp     = array[i];
    array[i] = array[numElements - i - 1];
    array[numElements - i - 1] = temp;
  }
 
}

void debugNeighbors(vector<int> neighborIndices, int soughtIndex) {
    
  unsigned int j;
  bool debugOK = false;

  for (j = 0; j < neighborIndices.size(); j++) {
    if (neighborIndices[j] == soughtIndex) {
      debugOK = true;
    }
  }
  if (!debugOK) {
    cout << "Neighbors does'nt include soughtindex = " <<  soughtIndex << endl;
    assert(0);
  }
}


///
void findNeighbors (vector<int> &neighborIndices, int startNode, 
		    Protein ** allNodes) 
{
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  while (linkPtr != NULL) 
    {
      neighborIndices.push_back(linkPtr->node->myConfig->getIndex());
      linkPtr                     = linkPtr->nextPtr;
    }
}

/*
  finds and returns both node indices that the input node is connected
  to, as well as the edge weights.
*/
void findNeighborsAndEdgeWeights (vector<int>   &neighborIndices,
				  vector<float> &edgeWeights,
				  int startNode,
				  Protein ** allNodes)
{
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  edgeWeights    .resize(0);

  while (linkPtr != NULL) 
    {
      neighborIndices.push_back (linkPtr->node->myConfig->getIndex());
      edgeWeights    .push_back (linkPtr->pathWeight);
      linkPtr       = linkPtr->nextPtr;
    }
}

/*
  finds and returns both node indices that the input node is connected
  to, as well as the edge weights.
*/
void findNeighborsAndTransitionProbabilities (vector<int>   &neighborIndices,
					      vector<float> &tranProbs,
					      int startNode,
					      Protein ** allNodes)
{
  ptrList * linkPtr;
  
  linkPtr  = allNodes[startNode]->myNode->firstLink;

  neighborIndices.resize(0);
  tranProbs    .resize(0);

  while (linkPtr != NULL) 
    {
      neighborIndices.push_back (linkPtr->node->myConfig->getIndex());
      tranProbs      .push_back (linkPtr->edgeTransitionProbability);
      linkPtr       = linkPtr->nextPtr;
    }
}


/*  allocates space, and sets its values to false. */
void prepareIncrementedArray(bool ** incremented, int numNodes) {
  (*incremented) = new bool[numNodes];
  for (int i = 0; i < numNodes; i++) {
    (*incremented) [i] = false;
  }
}
//@}
