#ifndef  _TREE_H
#define  _TREE_H

#include "protein.H"
#include "analysis.H"
//#include "pairOfVector.H"

class ptrList;
class nodeClass;
class treeClass;
class Protein;

class ptrList 
{
public:
  ptrList   *nextPtr;                   //points to the next edge from current node
  nodeClass *node;                      //points to the node that corresponds to this edge
  float      pathWeight;                // pathweight of current edge
  float      edgeTransitionProbability; // transition probability of current edge
  ptrList();
};

class treeClass 
{
public:
  nodeClass *root;      // a node in the connected component
  treeClass *nextTree;  //link that connect this connected component to the list of connected components
  treeClass *prevTree;  //same as nextTree
  int        numNodesInTree; // num nodes in the connected component
  int        myTreeName; //the distinctive id of the connected component

  treeClass();
  void removeTree (int *, treeClass* ,treeClass ** , 
		   treeClass **,Protein **, const int);
  void printTree  ();

  /* prints the indices of the nodes in the tree */
  // void printNodeIndices ();
};

struct queStruct {

  struct queStruct *nextQEntry;
  float timeStamp;
  nodeClass *fromNode;
  nodeClass *toNode;
};

typedef struct queStruct queType;

class nodeClass {

public:
  ptrList *firstLink; //points to first edge
  ptrList *lastLink;  //points to last edge
  Protein *myConfig;  //points to the node
  nodeClass *pathPtr;
  char nodeColor;     // used in graph traversal
  float nodeTime;

  nodeClass            ();
  void  removeAllLinks();
  void  addNodeLink     (nodeClass*, float);
  void  replaceTreePtr  (treeClass*, nodeClass*);
  void  setPath         (queType  *, nodeClass*, float);
  void  setPath_reverse (queType  *, nodeClass*, float);
  float reverseWeight   (nodeClass*);

};

class Path {
public :

  Path(int i, int j, int numberOfNodes, Protein ** allNodes); 
/* constructs the first node of the tree, 
			 which has the best path from node i to node j.
		      */
  Path(int deviatedEdgeIndex);
  /* 
     almost same as above.
     except : start and endIndex are now known.
     deviatedEdgeIndex set by user.
  */
  Path();                      // same as above. everything set by default.
    

  Path (float _pathWeight, int _numIndices, int * _indices,
	Protein ** allNodes);


  ~Path();

  void expand (int numberOfNodes, Protein ** allNodes);
  //  void appendNode(int nodeIndex); not used as of now
  void appendPath         (Path * path2Append,Protein ** allNodes);
  void copyPath  (Path * headPath);
  void copyPath  (int * indices, int numIndices ,Protein ** allNodes);
  //  void print(Protein ** allNodes,pairOfVector ** pairsOfVectors);
  /*
  generates atom coords on the backbone for each node on the path, 
  if not already generated, and then prints this to a file.
*/
  void print(Protein ** allNodes, const char * filename);
  void print(Protein ** allNodes);
  void appendEnergies2File(Protein ** allNodes, const char * filename);
  /*
    first prints numIndices, # parameters, 
    then prints parameters themselves.
    Between each set of parameters corresponding to a node, prints #.
  */
  
  void appendConfigParams2File (Protein ** allNodes, 
				const char * fileName);

  void append2Children(int numNewChildren, Path * newChildren, 
		   int inp_deviatedEdgeIndex);
  void findMoreDeviationsAndAddThemToChildren (const vector<int>&
					       neighborIndices,
					       int inp_deviatedEdgeIndex, 
					       int numberOfNodes, 
					       Protein ** allNodes);




  /* for debugging */
  
  friend void nextBestPath     (Path ** desiredPath, Path * root);
  friend void reportKBestPaths (int K, int startIndex, int endIndex, 
				int numberOfNodes);
  
  /* end for debugging */
  
  float  getPathWeight()  const;
  int    getNumChildren() const;
  Path * getChildPtr(int) const; 
  int    getIndex(int) const;
  int    getNumIndices() const;
  int    getDeviatedEdgeIndex() const;

  void   setDeviatedEdgeIndex(int);
  bool   isFound() const;
  void   setFound(); /* sets found to true */
  void   updatePathWeight(Protein ** allNodes); /* somehow path weihts computed by runbfs are not correct, this function corrects it. Also used by copypath version 2. */
  bool   checkPathWeights(Protein ** allNodes);
  static const int uninitializedDeviatedEdgeIndexValue;
  
  /* 
     For each node in the path, increments the
     corresponding element in the numVisited array. 
  */
  void markVisitedNodes(visitsClass * visits, int numNodes,
			float closeEnoughRMSD,  Protein ** allNodes);
 
private : 

  float pathWeight;
  int  * indices; /* indices[0] == startIndex for elements in the formed tree
		     indices[numIndices - 1] == endIndex ""
		     but not all elements are in the tree!
		   */
  int numIndices;
  static int startIndex, endIndex, numElementsInTree;
  Path * children;
  int numChildren;
  int deviatedEdgeIndex; /*  AtWhichThisPathDeviatesFromItsParent;
			     0 .. numEdges = numIndices - 1?
			 */
  bool found; /*  true if a path is among k best paths and 
		  has been found before.
		  
		  useful if a node is found 
		  but has no children, hence it is found again and again
	      */

};

void nextBestPath (int ** indices, int * numIndices);
// finds best path, then calls expand on that node for next run.


void KBestPaths();// calls nextBestPath K times. 
void findNeighbors (vector<int> &neighborIndices, int startNode, 
		    Protein ** allNodes);
/*
  finds and returns both node indices that the input node is connected
  to, as well as the edge weights.
*/
void findNeighborsAndEdgeWeights (vector<int>   &neighborIndices,
				  vector<float> &edgeWeights,
				  int startNode,
				  Protein ** allNodes);

void findNeighborsAndTransitionProbabilities (vector<int>   &neighborIndices,
					      vector<float> &tranProbs,
					      int startNode,
					      Protein ** allNodes);

void  reportKBestPaths (int K, int startIndex, int endIndex, int numberOfNodes, Protein ** allNodes);

/* finds the best path from node #i to node #j.
   returns the bestpath, which includes first and last indices.
   NULL if there is no path.

   hidden i/p : numberOfNodes, allNodes

*/
void findBestPath (Path ** bestPath, int startIndex, int endIndex, int numberOfNodes, Protein ** allNodes);

/*  allocates space, and sets its values to false. */
void prepareIncrementedArray(bool ** incremented, int numNodes);

/* 
   returns forward path weight between allNodes[startIndex] and 
   allNodes[endIndex].
 */
float fPathWeight(int startIndex, int endIndex, Protein ** allNodes);
#endif

