#ifndef VEC_UTILITY_H
#define VEC_UTILITY_H


#include <math.h>
#include "myVector.H"
#include <vector>
using std::vector;
#include <iostream>
using std::cout;
using std::endl;

typedef Point Vec;

/* dest = source */
void VecCopy(Vec dest, const Vec source);

/* computes A-B = Result */
inline void  VecSub(Vec Result, const Vec A, const Vec B);

void  VecScale (Vec Result, const float a, const Vec A);

inline float VecLength(const Vec P);

inline float VecDot (const Vec A, const Vec B);

void  VecAdd(Vec Result,const Vec A, const Vec B );

inline void  VecCross(Vec Result,const Vec A, const Vec B);

void  VecCombine(Vec Result,float a, const Vec A, float b, const Vec B);

inline int   VecEqual (const Vec A, const Vec B);

inline void  VecNormalize(Vec P);

inline void  obtainVecFromPoints(const Point& p_base, const Point& p_end, Vec& v);

void  matMult(mat4 myoutmat, const mat4 mymat1, const mat4 mymat2); 

//void mat3Mul(mat3 myoutmat, const mat3 mat1, const mat3 mat2);

// source of below function : www.eldeeb.com
float matrixInverse( float * pInverse, float * pMatrix, int iRank );

void  vec4Combine(vec4 D, float a, const vec4 unitV, const vec4 C);

void  vec4Scale(vec4 Result, float a, const vec4 V);

void  vec4Add (vec4 Result, const vec4 A, const vec4 B);

void  applyMatToVec4(vec4 outvec,  const mat4 themat, const vec4 invec);

void applyMatToVec4 (vector<float> &  outvec,      const mat4 themat, 
		     const vector<float> &  invec);
//void applyMatToVec3(vec3 outvec,  const mat3 themat, const vec3 invec);

void  vec4Sub( vec4 Result, const vec4 V1, const vec4 V2);

void  vec4Normalize(vec4);

float vec4Length(const vec4);

void  printPoint(Point A);

//void vec4Copy( vec4 dest, const vec4 source );

/* finds the fourth point in a protein backbone given the previous three atom coordinates, bond lengths, bond angles and dihedral angle. */
void   obtainFourthPoint(const Point& pA, const Point& pB, const Point& pC, Point  pD, float dihedral_angle, const float bond_angle, const float bond_length);

float  findRMSD(const Point p1[], const Point p2[], const int numAtoms);

double dihedral_angle (Point A, Point B, Point C, Point D);

/* given vectors P and Q, whose base and end points are given by P1,P2,Q1 and 
   Q2 respectively, this function finds the angle between these vectors in 
   radian and returns it.
*/
double findAngle (const Point p1, const Point p2, 
		  const Point q1, const Point q2);

/* at the end, either P has unit length or 0 length */
inline void VecNormalize(Vec P)
{
  float length = 0.;
  int i;
  
  length = VecLength(P);

  if (length != 0.0) {
    for (i=0; i<3; i++)
      P[i] /= length;
  }
  else {
    cout << "P had 0 length so couldn't normalize it. " << endl;
  }
}

inline float VecLength(const Vec P)
{
  float length = 0.;
  int i;
  
  for (i=0; i<3; i++)
    length += P[i] * P[i];

  length = sqrt(length);
  return length;
}

inline void  VecSub(Vec Result, const Vec A, const Vec B)
{
  for (int i=0; i<3; i++)
    Result[i] = A[i] - B[i];
}

inline float VecDot (const Vec A, const Vec B) {
  int i;
  float retVal=0.;

  for (i=0; i<3; i++)
    retVal += A[i] * B[i];
  return retVal;
}

/* component-wise multiplication */
inline void VecMul(Vec Result,const Vec A, const Vec B) 
{
  int i;
  for (i=0; i<3; i++) {
    Result[i] = A[i] * B[i];
  }
}

/* return the cross product Result = A cross B */
inline void VecCross(Vec Result,const Vec A, const Vec B) {

	Result[0] = (A[1]*B[2]) - (A[2]*B[1]);
	Result[1] = (A[2]*B[0]) - (A[0]*B[2]);
	Result[2] = (A[0]*B[1]) - (A[1]*B[0]);
}

inline int VecEqual (const Vec A, const Vec B) {

  if (A[0] != B[0]) return 0;
  if (A[1] != B[1]) return 0;
  if (A[2] != B[2]) return 0;
  return 1;
}

inline void  obtainVecFromPoints(const Point& p_base, const Point& p_end, Vec& v) {
  v[0] = p_end[0] - p_base[0];
  v[1] = p_end[1] - p_base[1];
  v[2] = p_end[2] - p_base[2];
}


#endif
